/*
 * Copyright (c) Meta Platforms, Inc. and affiliates.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <thrift/compiler/whisker/object.h>

#include <thrift/compiler/whisker/detail/overload.h>

#include <cassert>
#include <ostream>
#include <set>
#include <sstream>
#include <type_traits>

#include <fmt/core.h>
#include <fmt/ranges.h>

namespace whisker {

namespace {

class to_string_visitor {
 public:
  explicit to_string_visitor(const object_print_options& opts) : opts_(opts) {}

  // Prevent implicit conversion to whisker::object. Otherwise, we can silently
  // compile an infinitely recursive visit() chain if there is a missing
  // overload for one of the alternatives in the variant.
  template <
      typename T = object,
      typename = std::enable_if_t<std::is_same_v<T, object>>>
  void visit(const T& value, tree_printer::scope scope) const {
    value.visit(
        [&](const array& a) { visit_maybe_truncate(a, std::move(scope)); },
        [&](const map& m) { visit_maybe_truncate(m, std::move(scope)); },
        [&](const native_object::ptr& o) {
          visit_maybe_truncate(o, std::move(scope));
        },
        [&](const native_function::ptr& f) {
          visit_maybe_truncate(f, std::move(scope));
        },
        // All other types are printed inline so no truncation is necessary.
        [&](auto&& alternative) { visit(alternative, std::move(scope)); });
  }

 private:
  template <typename T>
  void visit_maybe_truncate(const T& value, tree_printer::scope scope) const {
    if (at_max_depth(scope)) {
      scope.println("...");
      return;
    }
    visit(value, scope);
  }

  void visit(i64 value, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("i64({})", value);
  }

  void visit(f64 value, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("f64({})", value);
  }

  void visit(const std::string& value, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("'{}'", tree_printer::escape(value));
  }

  void visit(bool value, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("{}", value ? "true" : "false");
  }

  void visit(null, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("null");
  }

  void visit(const array& arr, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("array (size={})", arr.size());
    for (std::size_t i = 0; i < arr.size(); ++i) {
      auto element_scope = scope.open_transparent_property();
      element_scope.println("[{}]", i);
      visit(arr[i], element_scope.open_node());
    }
  }

  void visit(const map& m, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    scope.println("map (size={})", m.size());
    for (const auto& [key, value] : m) {
      auto element_scope = scope.open_transparent_property();
      element_scope.println("'{}'", key);
      visit(value, element_scope.open_node());
    }
  }

  void visit(const native_object::ptr& o, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    o->print_to(std::move(scope), opts_);
  }

  void visit(const native_function::ptr& f, tree_printer::scope scope) const {
    require_within_max_depth(scope);
    f->print_to(std::move(scope), opts_);
  }

  [[nodiscard]] bool at_max_depth(const tree_printer::scope& scope) const {
    return scope.semantic_depth() == opts_.max_depth;
  }

  void require_within_max_depth(
      [[maybe_unused]] const tree_printer::scope& scope) const {
    assert(scope.semantic_depth() <= opts_.max_depth);
  }

  const object_print_options& opts_;
};

} // namespace

void native_object::map_like::default_print_to(
    std::string_view name,
    std::vector<std::string> property_names,
    tree_printer::scope scope,
    const object_print_options& options) const {
  assert(scope.semantic_depth() <= options.max_depth);
  const auto size = property_names.size();
  scope.println("{} (size={})", name, size);

  for (const std::string& key : property_names) {
    auto cached = lookup_property(key);
    assert(cached != nullptr);
    auto element_scope = scope.open_transparent_property();
    element_scope.println("'{}'", key);
    whisker::print_to(*cached, element_scope.open_node(), options);
  }
}

void native_object::array_like::default_print_to(
    std::string_view name,
    tree_printer::scope scope,
    const object_print_options& options) const {
  assert(scope.semantic_depth() <= options.max_depth);

  const auto sz = size();
  scope.println("{} (size={})", name, sz);
  for (std::size_t i = 0; i < sz; ++i) {
    auto element_scope = scope.open_transparent_property();
    element_scope.println("[{}]", i);
    whisker::print_to(*at(i), element_scope.open_node(), options);
  }
}

void native_object::print_to(
    tree_printer::scope scope, const object_print_options&) const {
  scope.println("<native_object>");
}

bool native_object::operator==(const native_object& other) const {
  return &other == this;
}

native_function::context::array_like::array_like(
    native_object::array_like::ptr&& arr)
    : which_(std::move(arr)) {}
native_function::context::array_like::array_like(managed_ptr<array>&& arr)
    : which_(std::move(arr)) {}

native_function::context::array_like::~array_like() noexcept = default;

native_function::context::array_like::array_like(const array_like&) = default;

native_function::context::array_like&
native_function::context::array_like::operator=(const array_like&) = default;

native_function::context::array_like::array_like(array_like&&) noexcept =
    default;

native_function::context::array_like&
native_function::context::array_like::operator=(array_like&&) noexcept =
    default;

std::size_t native_function::context::array_like::size() const {
  return detail::variant_match(
      which_, [](const auto& arr) -> std::size_t { return arr->size(); });
}

object::ptr native_function::context::array_like::at(std::size_t index) const {
  return detail::variant_match(
      which_,
      [&](const native_object::array_like::ptr& arr) -> object::ptr {
        return arr->at(index);
      },
      [&](const managed_ptr<array>& arr) -> object::ptr {
        return manage_derived_ref(arr, (*arr)[index]);
      });
}

native_function::context::map_like::map_like(native_object::map_like::ptr&& m)
    : which_(std::move(m)) {}
native_function::context::map_like::map_like(managed_ptr<map>&& m)
    : which_(std::move(m)) {}

/* static */ std::optional<native_function::context::array_like>
native_function::context::array_like::try_from(const object::ptr& o) {
  if (o->is_array()) {
    return array_like(manage_derived_ref(o, o->as_array()));
  }
  if (o->is_native_object()) {
    if (native_object::array_like::ptr arr =
            o->as_native_object()->as_array_like()) {
      return array_like(std::move(arr));
    }
  }
  return std::nullopt;
}

native_function::context::map_like::~map_like() noexcept = default;

native_function::context::map_like::map_like(const map_like&) = default;

native_function::context::map_like&
native_function::context::map_like::operator=(const map_like&) = default;

native_function::context::map_like::map_like(map_like&&) noexcept = default;

native_function::context::map_like&
native_function::context::map_like::operator=(map_like&&) noexcept = default;

object::ptr native_function::context::map_like::lookup_property(
    std::string_view identifier) const {
  return detail::variant_match(
      which_,
      [&](const native_object::map_like::ptr& m) -> object::ptr {
        return m->lookup_property(identifier);
      },
      [&](const managed_ptr<map>& m) -> object::ptr {
        if (auto it = m->find(identifier); it != m->end()) {
          return manage_as_static(it->second);
        }
        return nullptr;
      });
}

std::optional<std::vector<std::string>>
native_function::context::map_like::keys() const {
  using result = std::optional<std::vector<std::string>>;
  return detail::variant_match(
      which_,
      [&](const native_object::map_like::ptr& m) -> result {
        return m->keys();
      },
      [&](const managed_ptr<map>& m) -> result {
        std::vector<std::string> keys;
        keys.reserve(m->size());
        for (const auto& [key, _] : *m) {
          keys.push_back(key);
        }
        return keys;
      });
}

/* static */ std::optional<native_function::context::map_like>
native_function::context::map_like::try_from(const object::ptr& o) {
  if (o->is_map()) {
    return map_like(manage_derived_ref(o, o->as_map()));
  }
  if (o->is_native_object()) {
    if (native_object::map_like::ptr m = o->as_native_object()->as_map_like()) {
      return map_like(std::move(m));
    }
  }
  return std::nullopt;
}

object::ptr native_function::context::named_argument(
    std::string_view name, named_argument_presence presence) const {
  auto arg = named_args_.find(name);
  if (arg == named_args_.end()) {
    if (presence == named_argument_presence::optional) {
      return nullptr;
    }
    error("Missing named argument '{}'.", name);
  }
  return arg->second;
}

void native_function::context::do_warning(std::string msg) const {
  diags_.get().warning(loc_.begin, "{}", std::move(msg));
}

void native_function::context::declare_arity(std::size_t expected) const {
  if (arity() != expected) {
    error("Expected {} argument(s) but got {}", expected, arity());
  }
}

void native_function::context::declare_named_arguments(
    std::initializer_list<std::string_view> expected) const {
  // Using std::set so that the error message is deterministic.
  std::set<std::string_view> names;
  for (const auto& [name, _] : named_args_) {
    names.insert(name);
  }
  for (const auto& name : expected) {
    names.erase(name);
  }
  if (!names.empty()) {
    error("Unknown named argument(s) provided: {}.", fmt::join(names, ", "));
  }
}

void native_function::print_to(
    tree_printer::scope scope, const object_print_options&) const {
  scope.println("<native_function>");
}

std::string to_string(const object& obj, const object_print_options& options) {
  std::ostringstream out;
  print_to(obj, tree_printer::scope::make_root(out), options);
  return std::move(out).str();
}

void print_to(
    const object& obj,
    tree_printer::scope scope,
    const object_print_options& options) {
  to_string_visitor(options).visit(obj, std::move(scope));
}

std::ostream& operator<<(std::ostream& out, const object& o) {
  return out << to_string(o);
}

} // namespace whisker
